#  Copyright (c) 2021 zfit

# TODO: improve errors of models. Generate more general error, inherit and use more specific?
import warnings


class PDFCompatibilityError(Exception):
    pass


class LogicalUndefinedOperationError(Exception):
    pass


class OperationNotAllowedError(Exception):
    pass


class ExtendedPDFError(Exception):
    pass


class AlreadyExtendedPDFError(ExtendedPDFError):
    pass


class NotExtendedPDFError(ExtendedPDFError):
    pass


class ConversionError(Exception):
    pass


class SubclassingError(Exception):
    pass


class BasePDFSubclassingError(SubclassingError):
    pass


class MinimizerSubclassingError(SubclassingError):
    pass


class IntentionAmbiguousError(Exception):
    pass


class UnderdefinedError(IntentionAmbiguousError):
    pass


class LimitsUnderdefinedError(UnderdefinedError):
    pass


class NormRangeUnderdefinedError(UnderdefinedError):
    pass


class OverdefinedError(IntentionAmbiguousError):
    pass


class LimitsOverdefinedError(OverdefinedError):
    pass


class CoordinatesUnderdefinedError(UnderdefinedError):
    pass


class AxesAmbiguousError(IntentionAmbiguousError):
    pass


class NotSpecifiedError(Exception):
    pass


class LimitsNotSpecifiedError(NotSpecifiedError):
    pass


class NormRangeNotSpecifiedError(NotSpecifiedError):
    pass


class AxesNotSpecifiedError(NotSpecifiedError):
    pass


class ObsNotSpecifiedError(NotSpecifiedError):
    pass


# Parameter Errors
class NameAlreadyTakenError(Exception):
    pass


# Operation errors
class IncompatibleError(Exception):
    pass


class ShapeIncompatibleError(IncompatibleError):
    pass


class ObsIncompatibleError(IncompatibleError):
    pass


class AxesIncompatibleError(IncompatibleError):
    pass


class CoordinatesIncompatibleError(IncompatibleError):
    pass


class SpaceIncompatibleError(IncompatibleError):
    pass


class LimitsIncompatibleError(IncompatibleError):
    pass


class NumberOfEventsIncompatibleError(ShapeIncompatibleError):
    pass


class InvalidLimitSubspaceError(Exception):
    pass


class ModelIncompatibleError(IncompatibleError):
    pass


# Data errors
class WeightsNotImplementedError(Exception):
    pass


class DataIsBatchedError(Exception):
    pass


# Parameter errors
class ParameterNotIndependentError(Exception):
    pass


# Minimizer errors

class NotMinimizedError(Exception):
    pass


# Runtime Errors


class IllegalInGraphModeError(Exception):
    pass


class CannotConvertToNumpyError(Exception):
    pass


# Baseclass to steer execution
class ZfitNotImplementedError(NotImplementedError):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        if type(self) == ZfitNotImplementedError:
            warnings.warn("Prefer to use a more specific subclass. See in `zfit.exceptions`")


class FunctionNotImplemented(ZfitNotImplementedError):
    """Any function, e.g. in a BaseModel, that not implemented and a fallback should be called.

    Preferably use more specific exceptions
    """


class StandardControlFlow(Exception):
    """An exception that inherits from this class will be regarded as part of the standard control flow and not as an
    Error.

    For example, if a function raises that values are NaN, this is often intercepted on purpose.
    """


class SpecificFunctionNotImplemented(FunctionNotImplemented):
    """If a specific function, e.g. by the user is not implemented."""


class MinimizeNotImplemented(FunctionNotImplemented):
    """The `minimize` function of a minimizer is not implemented."""


class MinimizeStepNotImplemented(FunctionNotImplemented):
    """The `step` function of a minimizer is not implemented."""


class AnalyticNotImplemented(ZfitNotImplementedError):
    """General exception if an analytic way is not implemented."""


class AnalyticIntegralNotImplemented(AnalyticNotImplemented):
    """If an analytic integral is not provided."""
    pass


class AnalyticSamplingNotImplemented(AnalyticNotImplemented):
    """If analytic sampling from a distribution is not possible."""
    pass


# PDF class internal handling errors
class NormRangeNotImplemented(StandardControlFlow):
    """Indicates that a function does not support the normalization range argument `norm_range`."""
    pass


class MultipleLimitsNotImplemented(StandardControlFlow):
    """Indicates that a function does not support several limits in a :py:class:`~zfit.Space`."""
    pass

class InitNotImplemented(StandardControlFlow):
    """Indicates that a minimize method does not support a FitResult instead of a loss."""
    pass


class VectorizedLimitsNotImplemented(StandardControlFlow):
    """Indicates that a function does not support vectorized (n_events > 1) limits in a :py:class:`~zfit.Space`."""
    pass


# Developer verbose messages

class WorkInProgressError(Exception):
    """Only for developing purpose!

    Does not serve as a 'real' Exception.
    """
    pass


class BreakingAPIChangeError(Exception):
    def __init__(self, msg, *args: object) -> None:
        default_msg = ("This item has been removed due to an API change. Instruction to update:\n"
                       "")
        msg = default_msg + str(msg)
        super().__init__(msg, *args)


class BehaviorUnderDiscussion(Exception):

    def __init__(self, msg, *args: object) -> None:
        default_msg = ("The behavior of the following is currently under discussion and ideas are well needed. "
                       "Please open an issue at https://github.com/zfit/zfit/issues with your opinion about this.\n"
                       "")
        msg = default_msg + str(msg)
        super().__init__(msg, *args)


class MaximumIterationReached(StandardControlFlow):
    pass



#  Copyright (c) 2021 zfit
import copy

import numpy as np
import pytest
import tensorflow as tf

import zfit

obs1_random = zfit.Space(obs="obs1", limits=(-1.5, 1.2))
obs1 = zfit.Space(obs="obs1", limits=(-1, 1))

coeffs_parametrization = [
    1.4,
    [0.6],
    [1.42, 1.2],
    [0.2, 0.8, 0.5],
    [8.1, 1.4, 3.6, 4.1],
    [1.1, 1.42, 1.2, 0.4, 0.7],
    [11.1, 1.4, 5.6, 3.1, 18.1, 3.1],
]

rel_integral = 7e-2
default_sampling = 100000

poly_pdfs = [(zfit.pdf.Legendre, default_sampling),
             (zfit.pdf.Chebyshev, default_sampling),
             (zfit.pdf.Chebyshev2, default_sampling),
             (zfit.pdf.Hermite, default_sampling * 20),
             (zfit.pdf.Laguerre, default_sampling * 20)]


@pytest.mark.parametrize("poly_cfg", poly_pdfs)
@pytest.mark.parametrize("coeffs", coeffs_parametrization)
@pytest.mark.flaky(3)
def test_polynomials(poly_cfg, coeffs):
    coeffs = copy.copy(coeffs)
    poly_pdf, n_sampling = poly_cfg
    polynomial = poly_pdf(obs=obs1, coeffs=coeffs)
    polynomial2 = poly_pdf(obs=obs1, coeffs=coeffs)

    polynomial_coeff0 = poly_pdf(obs=obs1, coeffs=coeffs, coeff0=1.)
    lower, upper = obs1.rect_limits
    x = np.random.uniform(size=(1000,), low=lower[0], high=upper[0])
    y_poly = polynomial.pdf(x)
    y_poly_u = polynomial.pdf(x, norm_range=False)
    y_poly2 = polynomial2.pdf(x)
    y_poly2_u = polynomial2.pdf(x, norm_range=False)
    y_poly_coeff0 = polynomial_coeff0.pdf(x)
    y_poly_coeff0_u = polynomial_coeff0.pdf(x, norm_range=False)
    y_poly_np, y_poly2_np, y_poly_coeff0_np = [y_poly.numpy(), y_poly2.numpy(), y_poly_coeff0.numpy()]
    y_polyu_np, y_poly2u_np, y_polyu_coeff0_np = [y_poly_u.numpy(), y_poly2_u.numpy(), y_poly_coeff0_u.numpy()]
    np.testing.assert_allclose(y_polyu_np, y_poly2u_np)
    np.testing.assert_allclose(y_polyu_np, y_polyu_coeff0_np)
    np.testing.assert_allclose(y_poly_np, y_poly2_np)
    np.testing.assert_allclose(y_poly_np, y_poly_coeff0_np)

    # test 1 to 1 range
    integral = polynomial.analytic_integrate(limits=obs1, norm_range=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs1, norm_range=False)
    analytic_integral = integral.numpy()
    assert pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral.numpy()

    # test with different range scaling
    polynomial = poly_pdf(obs=obs1_random, coeffs=coeffs)

    # test with limits != space
    integral = polynomial.analytic_integrate(limits=obs1, norm_range=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs1, norm_range=False)
    analytic_integral = integral.numpy()
    assert pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral.numpy()

    # test with limits == space
    integral = polynomial.analytic_integrate(limits=obs1_random, norm_range=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs1_random, norm_range=False)
    analytic_integral = integral.numpy()
    assert pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral.numpy()

    lower, upper = obs1_random.limit1d
    sample = tf.random.uniform((n_sampling, 1), lower, upper, dtype=tf.float64)
    test_integral = np.average(polynomial.pdf(sample, norm_range=False)) * obs1_random.rect_area()
    assert pytest.approx(analytic_integral, rel=rel_integral * 3) == test_integral



#  Copyright (c) 2021 zfit

from typing import Callable, Iterable, List, Optional, Tuple, Union

import tensorflow as tf
from tensorflow_probability import distributions as tfd

import zfit.z.numpy as znp

from .. import settings, z
from ..settings import run, ztypes
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import WorkInProgressError
from .data import Data
from .interfaces import ZfitPDF, ZfitSpace
from .space import Space


class UniformSampleAndWeights:
    def __call__(self, n_to_produce: Union[int, tf.Tensor], limits: Space, dtype):
        rnd_samples = []
        thresholds_unscaled_list = []
        weights = tf.broadcast_to(z.constant(1., shape=(1,)), shape=(n_to_produce,))
        n_produced = tf.constant(0, tf.int64)
        for i, space in enumerate(limits):
            lower, upper = space.rect_limits  # TODO: remove new space
            if i == len(limits) - 1:
                n_partial_to_produce = n_to_produce - n_produced  # to prevent roundoff errors, shortcut for 1 space
            else:
                if isinstance(space, EventSpace):
                    frac = 1.  # TODO(Mayou36): remove hack for Eventspace
                else:
                    tot_area = limits.rect_area()
                    frac = (space.rect_area() / tot_area)[0]
                n_partial_to_produce = tf.cast(
                    z.to_real(n_to_produce) * z.to_real(frac), dtype=tf.int64)  # TODO(Mayou36): split right!

            sample_drawn = tf.random.uniform(shape=(n_partial_to_produce, limits.n_obs + 1),
                                             # + 1 dim for the function value
                                             dtype=ztypes.float)

            rnd_sample = sample_drawn[:, :-1] * (upper - lower) + lower  # -1: all except func value
            thresholds_unscaled = sample_drawn[:, -1]
            # if not multiple_limits:
            #     return rnd_sample, thresholds_unscaled
            rnd_samples.append(rnd_sample)
            thresholds_unscaled_list.append(thresholds_unscaled)
            n_produced += n_partial_to_produce

        rnd_sample = znp.concatenate(rnd_samples, axis=0)
        thresholds_unscaled = znp.concatenate(thresholds_unscaled_list, axis=0)

        n_drawn = n_to_produce
        return rnd_sample, thresholds_unscaled, weights, weights, n_drawn


class EventSpace(Space):
    """EXPERIMENTAL SPACE CLASS!"""

    def __init__(self, obs: ztyping.ObsTypeInput, limits: ztyping.LimitsTypeInput, factory=None, dtype=ztypes.float,
                 name: Optional[str] = "Space"):
        if limits is None:
            raise ValueError("Limits cannot be None for EventSpaces (currently)")
        self._limits_tensor = None
        self.dtype = dtype
        self._factory = factory
        super().__init__(obs, limits, name)

    @property
    def factory(self):
        return self._factory

    @property
    def is_generator(self):
        return self.factory is not None

    @property
    def limits(self) -> ztyping.LimitsTypeReturn:
        limits = super().limits
        limits_tensor = self._limits_tensor
        if limits_tensor is not None:
            lower, upper = limits
            new_bounds = [[], []]
            for i, old_bounds in enumerate(lower, upper):
                for bound in old_bounds:
                    if self.is_generator:
                        new_bound = tuple(lim(limits_tensor) for lim in bound)
                    else:
                        new_bound = tuple(lim() for lim in bound)
                    new_bounds[i].append(new_bound)
                new_bounds[i] = tuple(new_bounds[i])
            limits = tuple(new_bounds)

        return limits

    def create_limits(self, n):
        if self._factory is not None:
            self._limits_tensor = self._factory(n)

    def iter_areas(self, rel: bool = False) -> Tuple[float, ...]:
        if not rel:
            raise RuntimeError("Currently, only rel with one limits is implemented in EventSpace")
        return (1.,)  # TODO: remove HACK, use tensors?

    def add(self, other: ztyping.SpaceOrSpacesTypeInput):
        raise RuntimeError("Cannot be called with an event space.")

    def combine(self, other: ztyping.SpaceOrSpacesTypeInput):
        raise RuntimeError("Cannot be called with an event space.")

    @staticmethod
    def _calculate_areas(limits) -> Tuple[float]:
        # TODO: return the area as a tensor?
        return (1.,)

    def __hash__(self):
        return id(self)


# TODO: estimate maximum initially?
@z.function(wraps='sample')
def accept_reject_sample(prob: Callable, n: int, limits: ZfitSpace,
                         sample_and_weights_factory: Callable = UniformSampleAndWeights,
                         dtype=ztypes.float, prob_max: Union[None, int] = None,
                         efficiency_estimation: float = 0.5) -> tf.Tensor:
    """Accept reject sample from a probability distribution.

    Args:
        prob: A function taking x a Tensor as an argument and returning the probability
            (or anything that is proportional to the probability).
        n: Number of samples to produce
        limits: The limits to sample from
        sample_and_weights_factory: An (immutable!) factory function that returns the following function:
            A function that returns the sample to insert into `prob` and the weights
            (probability density) of each sample together with the random thresholds. The API looks as follows:

            - Parameters:

                - n_to_produce (int, tf.Tensor): The number of events to produce (not exactly).
                - limits (Space): the limits in which the samples will be.
                - dtype (dtype): DType of the output.

            - Return:
                A tuple of length 5:
                - proposed sample (tf.Tensor with shape=(n_to_produce, n_obs)): The new (proposed) sample
                    whose values are inside `limits`.
                - thresholds_unscaled (tf.Tensor with shape=(n_to_produce,): Uniformly distributed
                    random values **between 0 and 1**.
                - weights (tf.Tensor with shape=(n_to_produce)): (Proportional to the) probability
                    for each sample of the distribution it was drawn from.
                - weights_max (int, tf.Tensor, None): The maximum of the weights (if known). This is
                    what the probability maximum will be scaled with, so it should be rather lower than the maximum
                    if the peaks do not exactly coincide. Otherwise return None (which will **assume**
                    that the peaks coincide).
                - n_produced: the number of events produced. Can deviate from the requested number.

        dtype:
        prob_max: The maximum of the model function for the given limits. If None
            is given, it will be automatically, safely estimated (by a 10% increase in computation time
            (constant weak scaling)).
        efficiency_estimation: estimation of the initial sampling efficiency.

    Returns:
    """
    prob_max_init = prob_max
    multiple_limits = len(limits) > 1
    if prob_max is None:
        n_min_to_produce = 3000
        overestimate_factor_scaling = 1.25
    else:  # if an exact estimation is given
        n_min_to_produce = 0
        overestimate_factor_scaling = 1.

    sample_and_weights = sample_and_weights_factory()
    n = tf.cast(n, dtype=tf.int64)
    if run.numeric_checks:
        tf.debugging.assert_non_negative(n)

    # whether we may produce more then n, we normally do (except for EventSpace which is not a generator)
    # we cannot cut inside the while loop as soon as we have produced enough because we may sample from
    # multiple limits and therefore need to randomly remove events, otherwise we are biased because the
    # drawn samples are ordered in the different
    dynamic_array_shape = True

    # for fixed limits in EventSpace we need to know which indices have been successfully sampled. Therefore this
    # can be None (if not needed) or a boolean tensor with the size `n`
    initial_is_sampled = tf.constant("EMPTY")
    if (isinstance(limits, EventSpace) and not limits.is_generator) or limits.n_events > 1:
        dynamic_array_shape = False
        if run.numeric_checks:
            tf.debugging.assert_equal(tf.cast(limits.n_events, dtype=tf.int64), n)

        initial_is_sampled = tf.fill(value=False, dims=(n,))
        efficiency_estimation = 1.0  # generate exactly n

    inital_n_produced = tf.constant(0, dtype=tf.int64)
    initial_n_drawn = tf.constant(0, dtype=tf.int64)
    sample = tf.TensorArray(dtype=dtype, size=tf.cast(n, dtype=tf.int32),
                            dynamic_size=dynamic_array_shape,
                            clear_after_read=True,  # we read only once at end to tensor
                            element_shape=(limits.n_obs,))

    @z.function(wraps='tensor')
    def not_enough_produced(n, sample, n_produced, n_total_drawn, eff, is_sampled, weights_scaling,
                            weights_maximum, prob_maximum, n_min_to_produce):
        return tf.greater(n, n_produced)

    @z.function(wraps='tensor')
    def sample_body(n, sample, n_produced=0, n_total_drawn=0, eff=1.0, is_sampled=None, weights_scaling=0.,
                    weights_maximum=0., prob_maximum=0., n_min_to_produce=10000):
        eff = znp.max([eff, z.to_real(1e-6)])
        n_to_produce = n - n_produced

        if isinstance(limits, EventSpace):  # EXPERIMENTAL(Mayou36): added to test EventSpace
            limits.create_limits(n=n)

        do_print = settings.get_verbosity() > 5
        if do_print:
            tf.print("Number of samples to produce:", n_to_produce, " with efficiency ", eff,
                     " with total produced ", n_produced, " and total drawn ", n_total_drawn,
                     " with weights scaling", weights_scaling)

        if dynamic_array_shape:
            # TODO: move all this fixed numbers out into settings
            # this section is non-trivial due to numerical issues with large numbers in n_to_produce
            # and small numbers in the efficiency. To prevent an overflow, we scale the efficieny up
            # and convert it to an int, effectively precision after the floaning point.
            # Then we scale it down again.
            eff_precision: int = 100
            one_over_eff_int = tf.cast(1 / eff * 1.01 * eff_precision, dtype=tf.int64)
            n_to_produce *= one_over_eff_int
            n_to_produce = znp.floor_divide(n_to_produce, eff_precision)
            # tf.debugging.assert_positive(n_to_produce_float, "n_to_produce went negative, overflow?")
            # n_to_produce = tf.cast(n_to_produce_float, dtype=tf.int64) + 3  # just to make sure
            n_to_produce = znp.maximum(n_to_produce, n_min_to_produce)
            # TODO: adjustable efficiency cap for memory efficiency (prevent too many samples at once produced)
            max_produce_cap = tf.constant(800000, dtype=tf.int64)
            tf.debugging.assert_positive(n_to_produce, "n_to_produce went negative, overflow?")
            # TODO: remove below? was there due to overflow in tf?
            # n_to_produce = znp.maximum(5, n_to_produce)  # protect against overflow, n_to_prod -> neg.
            n_to_produce = znp.minimum(n_to_produce, max_produce_cap)  # introduce a cap to force serial
            new_limits = limits  # because limits in the vector space case can change
        else:
            # TODO(Mayou36): add cap for n_to_produce here as well
            if multiple_limits:
                raise WorkInProgressError("Multiple limits for fixed event space not yet implemented")
            is_not_sampled = tf.logical_not(is_sampled)
            lower, upper = limits._rect_limits_tf
            lower = tf.boolean_mask(tensor=lower, mask=is_not_sampled)
            upper = tf.boolean_mask(tensor=upper, mask=is_not_sampled)
            new_limits = limits.with_limits(limits=(lower, upper))
            draw_indices = tf.where(is_not_sampled)

        rnd_sample, thresholds_unscaled, weights, weights_max, n_drawn = sample_and_weights(
            n_to_produce=n_to_produce,
            limits=new_limits,
            dtype=dtype)

        rnd_sample = Data.from_tensor(obs=new_limits, tensor=rnd_sample)
        n_drawn = tf.cast(n_drawn, dtype=tf.int64)
        if run.numeric_checks: tf.debugging.assert_non_negative(n_drawn)
        n_total_drawn += n_drawn
        probabilities = prob(rnd_sample)
        shape_rnd_sample = tf.shape(input=rnd_sample)[0]
        if run.numeric_checks:
            tf.debugging.assert_equal(tf.shape(input=probabilities), shape_rnd_sample)

        if prob_max_init is None or weights_max is None:  # TODO(performance): estimate prob_max, after enough estimations -> fix it?

            # safety margin, predicting future, improve for small samples?
            weights_maximum_new = znp.max(weights)
            weights_maximum = znp.maximum(weights_maximum, weights_maximum_new)
            # if run.numeric_checks:
            #     tf.debugging.assert_greater_equal(weights, weights_maximum * 1e-5, message="The ")
            weights_clipped = znp.maximum(weights, weights_maximum * 1e-5)
            # prob_weights_ratio = probabilities / weights
            prob_maximum_new = znp.max(probabilities)
            prob_maximum = znp.maximum(prob_maximum, prob_maximum_new)
            # prob_weights_ratio = probabilities / weights_clipped
            prob_weights_ratio_max = znp.max(probabilities / weights_clipped)
            max_prob_weights_ratio = znp.maximum(prob_maximum / weights_maximum, prob_weights_ratio_max)
            # clipping means that we don't scale more for a certain threshold
            # to properly account for very small numbers, the thresholds should be scaled to match the ratio
            # but if a weight of a sample is very low (compared to the other weights), this would force the acceptance
            # of other samples to decrease strongly. We introduce a cut here, meaning that any event with an acceptance
            # chance of less then 1 in ratio_threshold will be underestimated.
            # TODO(Mayou36): make ratio_threshold a global setting
            # max_prob_weights_ratio_clipped = znp.minimum(max_prob_weights_ratio,
            #                                             min_prob_weights_ratio * ratio_threshold)
            # max_prob_weights_ratio_clipped = max_prob_weights_ratio
            new_scaling_needed = max_prob_weights_ratio > weights_scaling
            old_scaling = weights_scaling
            weights_scaling = tf.cond(new_scaling_needed,
                                      lambda: max_prob_weights_ratio * overestimate_factor_scaling,
                                      lambda: weights_scaling)

            def calc_new_n_produced():
                n_produced_float = tf.cast(n_produced, dtype=ztypes.float)
                binomial = tfd.Binomial(n_produced_float, probs=old_scaling / weights_scaling)
                return tf.cast(tf.round(binomial.sample()), dtype=tf.int64)

            n_produced = tf.cond(new_scaling_needed,
                                 calc_new_n_produced,
                                 lambda: n_produced)
            # TODO: for fixed array shape?
            # weights_scaling = znp.maximum(max_prob_weights_ratio, weights_scaling)

            # weights_scaling = znp.maximum(weights_scaling, max_prob_weights_ratio_clipped * (1 + 1e-2))
        else:
            weights_scaling = prob_max / weights_max

        weights_scaled = (weights_scaling * weights
                          # * (1 + 1e-8)
                          )  # numerical epsilon
        random_thresholds = thresholds_unscaled * weights_scaled
        if run.numeric_checks:
            invalid_probs_weights = tf.greater(probabilities, weights_scaled)
            failed_weights = tf.boolean_mask(tensor=weights_scaled, mask=invalid_probs_weights)

            # def bias_print():
            #     tf.print("HACK WARNING: if the following is NOT empty, your sampling _may_ be biased."
            #              " Failed weights:", failed_weights, " failed probs", failed_probs)

            # tf.cond(tf.not_equal(tf.shape(input=failed_weights), [0]), bias_print, lambda: None)

            tf.debugging.assert_equal(tf.shape(input=failed_weights), [0])

            # for weights scaled more then ratio_threshold
            if run.numeric_checks:
                tf.debugging.assert_greater_equal(x=weights_scaled, y=probabilities,
                                                  message="Not all weights are >= probs so the sampling "
                                                          "will be biased. If a custom `sample_and_weights` "
                                                          "was used, make sure that either the shape of the "
                                                          "custom sampler (resp. it's weights) overlap better "
                                                          "or decrease the `max_weight`")
            #
            # # check disabled (below not added to deps)
            # assert_scaling_op = tf.assert_less(weights_scaling / min_prob_weights_ratio, z.constant(ratio_threshold),
            #                                    data=[weights_scaling, min_prob_weights_ratio],
            #                                    message="The ratio between the probabilities from the pdf and the"
            #                                    f"probability from the sampler is higher "
            #                                    f" then {ratio_threshold}. This will most probably bias the sampling. "
            #                                    f"Use importance sampling or, to disable this check, do"
            #                                    f"zfit.run.numeric_checks = False")
            # assert_op.append(assert_scaling_op)

        take_or_not = probabilities > random_thresholds
        take_or_not = take_or_not[0] if len(take_or_not.shape) == 2 else take_or_not
        filtered_sample = tf.boolean_mask(tensor=rnd_sample, mask=take_or_not, axis=0)

        n_accepted = tf.shape(input=filtered_sample, out_type=tf.int64)[0]
        n_produced_new = n_produced + n_accepted
        if not dynamic_array_shape:
            indices = tf.boolean_mask(tensor=draw_indices, mask=take_or_not)
            current_sampled = tf.sparse.to_dense(tf.SparseTensor(indices=indices,
                                                                 values=tf.broadcast_to(input=(True,),
                                                                                        shape=(n_accepted,)),
                                                                 dense_shape=(tf.cast(n, dtype=tf.int64),)),
                                                 default_value=False)
            is_sampled = tf.logical_or(is_sampled, current_sampled)
            indices = indices[:, 0]
        else:
            indices = tf.range(n_produced, n_produced_new)

        # TODO: pack into tf.function to speedup considerable the eager sampling? Is bottleneck currently
        sample_new = sample.scatter(indices=tf.cast(indices, dtype=tf.int32), value=filtered_sample)

        # efficiency (estimate) of how many samples we get
        eff = znp.max([z.to_real(n_produced_new), z.to_real(1.)]) / znp.max(
            [z.to_real(n_total_drawn), z.to_real(1.)])
        return (
            n, sample_new, n_produced_new, n_total_drawn, eff, is_sampled, weights_scaling, weights_maximum,
            prob_maximum, n_min_to_produce
        )

    efficiency_estimation = z.to_real(efficiency_estimation)
    weights_scaling = z.constant(0.)
    weights_maximum = z.constant(0.)
    prob_maximum = z.constant(0.)
    n_min_to_produce = tf.cast(n_min_to_produce, dtype=tf.int64)
    inital_n_produced = tf.cast(inital_n_produced, dtype=tf.int64)
    initial_n_drawn = tf.cast(initial_n_drawn, dtype=tf.int64)

    loop_vars = (
        n, sample, inital_n_produced, initial_n_drawn,
        efficiency_estimation, initial_is_sampled, weights_scaling,
        weights_maximum, prob_maximum, n_min_to_produce)

    sample_array = tf.while_loop(cond=not_enough_produced, body=sample_body,  # paraopt
                                 loop_vars=loop_vars,
                                 swap_memory=True,
                                 parallel_iterations=1)[1]
    new_sample = sample_array.stack()
    new_sample = tf.stop_gradient(new_sample)  # stopping backprop

    if multiple_limits:
        new_sample = tf.random.shuffle(new_sample)  # to make sure, randomly remove and not biased.
    if dynamic_array_shape:  # if not dynamic we produced exact n -> no need to cut
        new_sample = new_sample[:n, :]  # cutting away to many produced

    # if no failure, uncomment both for improvement of shape inference, but what if n is tensor?
    # with suppress(AttributeError):  # if n_samples_int is not a numpy object
    #     new_sample.set_shape((n_samples_int, n_dims))
    return new_sample


def extract_extended_pdfs(pdfs: Union[Iterable[ZfitPDF], ZfitPDF]) -> List[ZfitPDF]:
    """Return all extended pdfs that are daughters.

    Args:
        pdfs:

    Returns:
        List[pdfs]:
    """
    from ..models.functor import BaseFunctor

    pdfs = convert_to_container(pdfs)
    indep_pdfs = []

    for pdf in pdfs:
        if not pdf.is_extended:
            continue
        elif isinstance(pdf, BaseFunctor):
            if all(pdf.pdfs_extended):
                indep_pdfs.extend(extract_extended_pdfs(pdfs=pdf.pdfs))
            elif not any(pdf.pdfs_extended):
                indep_pdfs.append(pdf)
            else:
                assert False, "Should not reach this point, wrong assumptions. Please report bug."
        else:  # extended, but not a functor
            indep_pdfs.append(pdf)

    return indep_pdfs


def extended_sampling(pdfs: Union[Iterable[ZfitPDF], ZfitPDF], limits: Space) -> tf.Tensor:
    """Create a sample from extended pdfs by sampling poissonian using the yield.

    Args:
        pdfs:
        limits:

    Returns:
        Union[tf.Tensor]:
    """
    samples = []
    pdfs = convert_to_container(pdfs)
    pdfs = extract_extended_pdfs(pdfs)

    for pdf in pdfs:
        n = tf.random.poisson(lam=pdf.get_yield(), shape=(), dtype=ztypes.float)
        n = tf.cast(n, dtype=tf.int64)
        sample = pdf._single_hook_sample(limits=limits, n=n)
        # sample.set_shape((n, limits.n_obs))
        samples.append(sample)

    samples = znp.concatenate(samples, axis=0)
    return samples



"""Functors are functions that take typically one or more other PDF. Prominent examples are a sum, convolution etc.

A FunctorBase class is provided to make handling the models easier.

Their implementation is often non-trivial.
"""
#  Copyright (c) 2021 zfit
import functools
import operator
from collections import OrderedDict
from typing import List, Optional

import tensorflow as tf

import zfit.z.numpy as znp

from .. import z
from ..core.basepdf import BasePDF
from ..core.coordinates import convert_to_obs_str
from ..core.interfaces import ZfitData, ZfitModel, ZfitPDF
from ..core.parameter import convert_to_parameter
from ..core.space import supports
from ..models.basefunctor import FunctorMixin, extract_daughter_input_obs
from ..settings import run, ztypes
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.exception import (AnalyticIntegralNotImplemented,
                              ModelIncompatibleError,
                              NormRangeUnderdefinedError, ObsIncompatibleError,
                              SpecificFunctionNotImplemented)
from ..util.warnings import warn_advanced_feature, warn_changed_feature
from ..z.random import counts_multinomial


class BaseFunctor(FunctorMixin, BasePDF):

    def __init__(self, pdfs, name="BaseFunctor", **kwargs):
        self.pdfs = convert_to_container(pdfs)
        super().__init__(models=self.pdfs, name=name, **kwargs)
        self._set_norm_range_from_daugthers()

    def _set_norm_range_from_daugthers(self):
        norm_range = super().norm_range
        if not norm_range.limits_are_set:
            norm_range = extract_daughter_input_obs(obs=norm_range,
                                                    spaces=[model.space for model in self.models])
        if not norm_range.limits_are_set:
            raise NormRangeUnderdefinedError(
                f"Daughter pdfs {self.pdfs} do not agree on a `norm_range` and/or no `norm_range`"
                "has been explicitly set.")

        self.set_norm_range(norm_range)

    @property
    def pdfs_extended(self):
        return [pdf.is_extended for pdf in self.pdfs]

    @property
    def _models(self) -> List[ZfitModel]:
        return self.pdfs


class SumPDF(BaseFunctor):

    def __init__(self, pdfs: List[ZfitPDF], fracs: Optional[ztyping.ParamTypeInput] = None,
                 obs: ztyping.ObsTypeInput = None,
                 name: str = "SumPDF"):
        """Create the sum of the `pdfs` with `fracs` as coefficients or the yields, if extended pdfs are given.

        If *all* pdfs are extended, the fracs is optional and the (normalized) yields will be used as fracs.
        If fracs is given, this will be used as the fractions, regardless of whether the pdfs have a yield or not.

        The parameters of the SumPDF are the fractions that are used to multiply the output of each daughter pdf.
        They can be accessed with `pdf.params` and have names f"frac_{i}" with i starting from 0 and going to the number
        of pdfs given.

        To get the component outputs of this pdf, e.g. to plot it, use `pdf.params.values()` to iterate through the
        fracs and `pdfs` to get the pdfs. For example

        .. code-block:: python

            for pdf, frac in zip(sumpdf.pdfs, sumpdf.params.values()):
                frac_integral = pdf.integrate(...) * frac

        Args:
            pdfs: The pdfs to be added.
            fracs: Coefficients for the linear combination of the pdfs. Optional if *all* pdfs are extended.

                - len(frac) == len(basic) - 1 results in the interpretation of a non-extended pdf.
                  The last coefficient will equal to 1 - sum(frac)
                - len(frac) == len(pdf): the fracs will be used as is and no normalization attempt is taken.
            name: |name_arg_descr|

        Raises
            ModelIncompatibleError: If model is incompatible.
        """
        # Check user input
        self._fracs = None

        pdfs = convert_to_container(pdfs)
        self.pdfs = pdfs
        if len(pdfs) < 2:
            raise ValueError(f"Cannot build a sum of a single pdf {pdfs}")
        common_obs = obs if obs is not None else pdfs[0].obs
        common_obs = convert_to_obs_str(common_obs)
        if not all(frozenset(pdf.obs) == frozenset(common_obs) for pdf in pdfs):
            raise ObsIncompatibleError("Currently, sums are only supported in the same observables")

        # check if all extended
        are_extended = [pdf.is_extended for pdf in pdfs]
        all_extended = all(are_extended)
        no_extended = not any(are_extended)

        fracs = convert_to_container(fracs)
        if fracs:  # not None or empty list
            fracs = [convert_to_parameter(frac) for frac in fracs]
        elif not all_extended:
            raise ModelIncompatibleError(f"Not all pdf {pdfs} are extended and no fracs {fracs} are provided.")

        if not no_extended and fracs:
            warn_advanced_feature(f"This SumPDF is built with fracs {fracs} and {'all' if all_extended else 'some'} "
                                  f"pdf are extended: {pdfs}."
                                  f" This will ignore the yields of the already extended pdfs and the result will"
                                  f" be a not extended SumPDF.", identifier='sum_extended_frac')

        # catch if args don't fit known case

        if fracs:
            # create fracs if one is missing
            if len(fracs) == len(pdfs) - 1:
                remaining_frac_func = lambda: tf.constant(1., dtype=ztypes.float) - tf.add_n(fracs)
                remaining_frac = convert_to_parameter(remaining_frac_func,
                                                      params=fracs)
                if run.numeric_checks:
                    tf.debugging.assert_non_negative(remaining_frac,
                                                     f"The remaining fraction is negative, the sum of fracs is > 0. Fracs: {fracs}")  # check fractions

                # IMPORTANT! Otherwise, recursion due to namespace capture in the lambda
                fracs_cleaned = fracs + [remaining_frac]

            elif len(fracs) == len(pdfs):
                warn_changed_feature("A SumPDF with the number of fractions equal to the number of pdf will no longer "
                                     "be extended. To make it extended, either manually use 'create_exteneded' or set "
                                     "the yield. OR provide all pdfs as extended pdfs and do not provide a fracs "
                                     "argument.", identifier='new_sum')
                fracs_cleaned = fracs

            else:
                raise ModelIncompatibleError(f"If all PDFs are not extended {pdfs}, the fracs {fracs} have to be of"
                                             f" the same length as pdf or one less.")
            param_fracs = fracs_cleaned

        # for the extended case, take the yields, normalize them, in case no fracs are given.
        if all_extended and not fracs:
            yields = [pdf.get_yield() for pdf in pdfs]

            def sum_yields_func():
                return znp.sum(
                    [tf.convert_to_tensor(value=y, dtype_hint=ztypes.float) for y in yields])

            sum_yields = convert_to_parameter(sum_yields_func, params=yields)
            yield_fracs = [convert_to_parameter(lambda sum_yields, yield_: yield_ / sum_yields,
                                                params=[sum_yields, yield_])
                           for yield_ in yields]

            fracs_cleaned = None
            param_fracs = yield_fracs

        self.pdfs = pdfs
        self._fracs = param_fracs
        self._original_fracs = fracs_cleaned

        params = OrderedDict()
        for i, frac in enumerate(param_fracs):
            params[f'frac_{i}'] = frac

        super().__init__(pdfs=pdfs, obs=obs, params=params, name=name)
        if all_extended and not fracs_cleaned:
            self.set_yield(sum_yields)
            # self.set_yield(sum_yields)  # TODO(SUM): why not the public method below?

    @property
    def fracs(self):
        return self._fracs

    def _apply_yield(self, value: float, norm_range: ztyping.LimitsType, log: bool):
        if all(self.pdfs_extended):
            return value
        else:
            return super()._apply_yield(value=value, norm_range=norm_range, log=log)

    def _unnormalized_pdf(self, x):  # NOT _pdf, as the normalization range can differ
        pdfs = self.pdfs
        fracs = self.params.values()
        probs = [pdf.pdf(x) * frac for pdf, frac in zip(pdfs, fracs)]
        prob = sum(probs)
        return z.convert_to_tensor(prob)

    def _pdf(self, x, norm_range):  # NOT _pdf, as the normalization range can differ
        equal_norm_ranges = len(set([pdf.norm_range for pdf in self.pdfs] + [norm_range])) == 1
        if not equal_norm_ranges:
            raise SpecificFunctionNotImplemented
        pdfs = self.pdfs
        fracs = self.params.values()
        probs = [pdf.pdf(x) * frac for pdf, frac in zip(pdfs, fracs)]
        prob = sum(probs)
        return z.convert_to_tensor(prob)

    @supports(multiple_limits=True)
    def _integrate(self, limits, norm_range):
        pdfs = self.pdfs
        fracs = self.fracs
        # TODO(SUM): why was this needed?
        # assert norm_range not in (None, False), "Bug, who requested an unnormalized integral?"
        integrals = [frac * pdf.integrate(limits=limits)  # do NOT propagate the norm_range!
                     for pdf, frac in zip(pdfs, fracs)]
        integral = sum(integrals)
        return z.convert_to_tensor(integral)

    @supports(multiple_limits=True)
    def _analytic_integrate(self, limits, norm_range):
        pdfs = self.pdfs
        fracs = self.fracs
        try:
            integrals = [frac * pdf.analytic_integrate(limits=limits)  # do NOT propagate the norm_range!
                         for pdf, frac in zip(pdfs, fracs)]
        except AnalyticIntegralNotImplemented as error:
            raise AnalyticIntegralNotImplemented(
                f"analytic_integrate of pdf {self.name} is not implemented in this"
                f" SumPDF, as at least one sub-pdf does not implement it.") from error

        integral = sum(integrals)
        return z.convert_to_tensor(integral)

    @supports(multiple_limits=True)
    def _partial_integrate(self, x, limits, norm_range):

        pdfs = self.pdfs
        fracs = self.fracs

        partial_integral = [pdf.partial_integrate(x=x, limits=limits) * frac  # do NOT propagate the norm_range!
                            for pdf, frac in zip(pdfs, fracs)]
        partial_integral = sum(partial_integral)
        return z.convert_to_tensor(partial_integral)

    @supports(multiple_limits=True)
    def _partial_analytic_integrate(self, x, limits, norm_range):
        pdfs = self.pdfs
        fracs = self.fracs
        try:
            partial_integral = [pdf.partial_analytic_integrate(x=x, limits=limits) * frac
                                # do NOT propagate the norm_range!
                                for pdf, frac in zip(pdfs, fracs)]
        except AnalyticIntegralNotImplemented as error:
            raise AnalyticIntegralNotImplemented(
                "partial_analytic_integrate of pdf {name} is not implemented in this"
                " SumPDF, as at least one sub-pdf does not implement it.") from error
        partial_integral = sum(partial_integral)
        return z.convert_to_tensor(partial_integral)

    @supports(multiple_limits=True)
    def _sample(self, n, limits):
        if (isinstance(n, str)):
            n = [n] * len(self.pdfs)
        else:
            n = tf.unstack(counts_multinomial(total_count=n, probs=self.fracs), axis=0)

        samples = []
        for pdf, n_sample in zip(self.pdfs, n):
            sub_sample = pdf.sample(n=n_sample, limits=limits)
            if isinstance(sub_sample, ZfitData):
                sub_sample = sub_sample.value()
            samples.append(sub_sample)
        sample = znp.concatenate(samples, axis=0)
        sample = tf.random.shuffle(sample)
        return sample


class ProductPDF(BaseFunctor):  # TODO: compose of smaller Product PDF by disassembling components subsets of obs
    def __init__(self, pdfs: List[ZfitPDF], obs: ztyping.ObsTypeInput = None, name="ProductPDF"):
        super().__init__(pdfs=pdfs, obs=obs, name=name)

    def _unnormalized_pdf(self, x: ztyping.XType):
        probs = [pdf.pdf(x, norm_range=False) for pdf in self.pdfs]
        prob = functools.reduce(operator.mul, probs)
        return z.convert_to_tensor(prob)

    def _pdf(self, x, norm_range):
        equal_norm_ranges = len(set([pdf.norm_range for pdf in self.pdfs] + [norm_range])) == 1  # all equal
        if not any(self._model_same_obs) and equal_norm_ranges:

            probs = [pdf.pdf(x=x) for pdf in self.pdfs]
            prob = functools.reduce(operator.mul, probs)
            return z.convert_to_tensor(prob)
        else:
            raise SpecificFunctionNotImplemented



#  Copyright (c) 2021 zfit

import pytest

import zfit
from zfit.core.dimension import limits_overlap, obs_subsets
from zfit.core.space import combine_spaces, limits_consistent
from zfit.util.exception import (LimitsNotSpecifiedError,
                                 MultipleLimitsNotImplemented,
                                 ObsIncompatibleError, SpaceIncompatibleError)

obs = ['obs' + str(i) for i in range(4)]
space1 = zfit.Space(obs=obs)
space2 = zfit.Space(obs=reversed(obs))
space3 = zfit.Space(obs=obs)
space4 = zfit.Space(obs=obs[0:2])

lower1d_1 = ((1,),)
upper1d_1 = ((2,),)

lower1d_12 = ((3,),)
upper1d_12 = ((4,),)

lower1d_22 = ((5,),)
upper1d_22 = ((6,),)

combined_lim_1d_12_and_22 = (((3,), (5,)), ((4,), (6,)))

lower1d_2 = ((1,),)
upper1d_2 = ((2,),)

# lower1d_2 = ((1,), (3,))
# upper1d_2 = ((2,), (4,))

lower2d_1 = ((1.5, 3),)
upper2d_1 = ((2, 4),)

lower2d_2 = ((1, 3),)
upper2d_2 = ((2, 4),)

space1d_1 = zfit.Space(obs=obs[0], limits=(lower1d_1, upper1d_1))
space1d_12 = zfit.Space(obs=obs[1], limits=(lower1d_12, upper1d_12))
space1d_22 = zfit.Space(obs=obs[1], limits=(lower1d_22, upper1d_22))
space1d_2 = zfit.Space(obs=obs[0], limits=(lower1d_2, upper1d_2))

space2d_1 = zfit.Space(obs=obs[:2], limits=(lower2d_1, upper2d_1))
space2d_2 = zfit.Space(obs=obs[:2], limits=(lower2d_2, upper2d_2))


def test_check_n_obs():
    with pytest.raises(SpaceIncompatibleError):
        zfit.pdf.Gauss(1., 4., obs=space2d_1)


def test_combine_spaces():
    combined_space = combine_spaces(space1d_1, space1d_12)
    combined_space2 = space1d_1 * space1d_12
    assert combined_space == space2d_2
    assert combined_space2 == space2d_2
    assert combine_spaces(space2d_2, space2d_2, space2d_2) == space2d_2
    none_limits_space = combine_spaces(space1, space1, space1)
    assert none_limits_space == space1  # with None limits
    assert not none_limits_space.limits_are_set
    with pytest.raises(LimitsNotSpecifiedError):
        combine_spaces(space2d_2, space1)

    # same but different syntax
    combined_space = space1d_1.combine(space1d_12)
    assert combined_space == space2d_2
    assert space2d_2.combine(space2d_2, space2d_2) == space2d_2
    none_limits_space = space1.combine(space1, space1)
    assert none_limits_space == space1  # with None limits
    assert not none_limits_space.limits_are_set
    with pytest.raises(LimitsNotSpecifiedError):
        space2d_2.combine(space1)


def test_add_spaces():
    from zfit.core.space import add_spaces as add_spaces

    assert add_spaces(space1, space2) == space1
    assert add_spaces(space1, space2, space3) == space1
    with pytest.raises(ObsIncompatibleError):
        add_spaces(space1, space2, space3, space4)

    assert space1 + space2 == space1
    assert space1 + space2 + space3 == space1
    with pytest.raises(ObsIncompatibleError):
        space1.add(space2, space3, space4)

    with pytest.raises(ObsIncompatibleError):
        add_spaces(space1d_2, space2d_1)
    with pytest.raises(MultipleLimitsNotImplemented):
        add_spaces(space1d_2, space1d_1).limits


def test_limits_consistent():
    assert limits_consistent(spaces=[space1, space2])
    assert limits_consistent(spaces=[space1, space2, space3])
    assert limits_consistent(spaces=[space1, space2, space3, space4])

    assert limits_consistent(spaces=[space1d_1, space1d_1])
    assert limits_consistent(spaces=[space1d_1, space1d_12])
    assert not limits_consistent(spaces=[space1d_1, space2d_1])
    assert limits_consistent(spaces=[space1d_1, space2d_2])
    assert limits_consistent(spaces=[space1d_1, space2d_2, space1d_12])
    assert not limits_consistent(spaces=[space1d_1, space2d_1, space1d_12])


def test_limits_overlap():
    assert not limits_overlap([space1d_1, space1d_2], allow_exact_match=True)
    assert limits_overlap([space1d_1, space1d_2])
    assert limits_overlap([space1d_1, space1d_2, space1d_12])
    assert limits_overlap([space1d_1, space2d_2, space1d_12])
    assert not limits_overlap([space1d_1, space2d_2, space1d_12], allow_exact_match=True)


def test_obs_subsets():
    space1 = zfit.Space('obs1')
    space2 = zfit.Space(['obs1', 'obs2', 'obs4'])
    space5 = zfit.Space(['obs3'])
    space6 = zfit.Space('obs6')
    space4 = zfit.Space(['obs3', 'obs7'])
    space3 = zfit.Space(['obs4', 'obs5'])
    spaces = [space1,
              space2,
              space5,
              space6,
              space4,
              space3]
    obs_subset = obs_subsets(spaces)

    obs1comb = frozenset(['obs1', 'obs2', 'obs4', 'obs5'])
    obs37comb = frozenset(['obs3', 'obs7'])
    obs6 = frozenset(['obs6'])
    true_obs = {obs1comb,
                obs37comb,
                obs6,
                }

    assert true_obs == set(obs_subset.keys())
    assert obs_subset[obs1comb] == [space1, space2, space3]
    assert obs_subset[obs37comb] == [space5, space4]
    assert obs_subset[obs6] == [space6]



"""Numpy like interface for math functions and arrays.

This module is intended to replace tensorflow specific methods and datastructures with equivalent or similar versions
in the numpy api. This should help make zfit as a project portable to alternatives of tensorflow should it be
necessary in the future.
At the moment it is simply an alias for the numpy api of tensorflow.
See https://www.tensorflow.org/guide/tf_numpy for more a guide to numpy api in tensorflow.
See https://www.tensorflow.org/api_docs/python/tf/experimental/numpy for the complete numpy api in tensorflow.

Recommended way of importing:
>>> Import zfit.z.numpy as znp
"""

from tensorflow.experimental.numpy import *



#  Copyright (c) 2021 zfit

from typing import List

import tensorflow as tf

from zfit.util.temporary import TemporarilySet


def all_parents(op, current_obs=None):
    if current_obs is None:
        current_obs = set()
    ops = {input_.op for input_ in op.inputs if input_.op not in current_obs}
    current_obs = current_obs.union(ops)
    return ops.union(*(all_parents(op, current_obs=current_obs) for op in ops))


def get_dependents_auto(tensor: tf.Tensor, candidates: List[tf.Tensor]) -> List[tf.Tensor]:
    """Return the nodes in `candidates` that `tensor` depends on.

    Args:
        tensor:
        candidates:
    """
    try:
        dependent_ops = all_parents(tensor.op)
    except RuntimeError as error:
        raise ValueError("Tensor too deeply nested, recursion limit exceeded. In the future,"
                         "implementation will be different and any dependents can be found."
                         "Currently, specify dependents explicitly if needed."
                         "Orignal Error: {}".format(error))
    dependent_candidates = [cand for cand in candidates if cand.op in dependent_ops]
    return dependent_candidates


class JIT:

    def _set_all(self, enable: bool = True):
        new_values = {k: enable for k in self._get_allowed()}

        def getter():
            return self._get_allowed().copy()

        def setter(jit_types):
            self._update_allowed(jit_types)

        return TemporarilySet(getter=getter, setter=setter, value=new_values)

    def _set_default(self):
        from zfit import z
        new_values = z.zextension.FunctionWrapperRegistry._DEFAULT_DO_JIT_TYPES.copy()

        for key in self._get_allowed():
            if key not in new_values:
                new_values[key] = new_values[key]  # default dict will explicitly set the default value

        def getter():
            return self._get_allowed().copy()

        def setter(jit_types):
            self._update_allowed(jit_types)

        return TemporarilySet(getter=getter, setter=setter, value=new_values)

    def _update_allowed(self, update_jit):
        from zfit import z

        z.zextension.FunctionWrapperRegistry.do_jit_types.update(update_jit)

    def _get_allowed(self):
        from zfit import z

        return z.zextension.FunctionWrapperRegistry.do_jit_types

    @property
    def experimental_is_eager(self):
        from ..settings import run
        return run.mode['graph']


jit = JIT()  # singleton



#  Copyright (c) 2021 zfit
import collections
import copy
import math
from typing import Callable, Mapping, Optional, Union

import nlopt
import numpy as np

from ..core.parameter import assign_values, set_values
from ..settings import run
from ..util.exception import MaximumIterationReached
from .baseminimizer import (NOT_SUPPORTED, BaseMinimizer, minimize_supports,
                            print_minimization_status)
from .fitresult import FitResult
from .strategy import ZfitStrategy
from .termination import CRITERION_NOT_AVAILABLE, EDM, ConvergenceCriterion


class NLoptBaseMinimizerV1(BaseMinimizer):
    _ALL_NLOPT_TOL = (
        # 'fatol',
        'ftol',
        'xatol',
        'xtol'
    )

    def __init__(self,
                 algorithm: int,
                 tol: Optional[float] = None,
                 gradient: Optional[Union[Callable, str, NOT_SUPPORTED]] = NOT_SUPPORTED,
                 hessian: Optional[Union[Callable, str, NOT_SUPPORTED]] = NOT_SUPPORTED,
                 maxiter: Optional[Union[int, str]] = None,
                 minimizer_options: Optional[Mapping[str, object]] = None,
                 internal_tols: Mapping[str, Optional[float]] = None,
                 verbosity: Optional[int] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt Base Minimizer V1"):
        """NLopt is a library that contains multiple different optimization algorithms.

         |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

         Args:

             algorithm: Define which algorithm to be used. These are taken from `nlopt.ALGORITHM` (where `ALGORITHM` is
                 the actual algorithm). A comprehensive list and description of all implemented algorithms is
                 available `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.
                 The wrapper is optimized for Local gradient-based optimization and may break with
                 others. However, please post a feature request in case other algorithms are requested.

                 The naming of the algorithm starts with either L/G (Local/Global)
                  and N/D (derivative-free/derivative-based).

                 Local optimizer options include (but are not limited to)

                 Derivative free:
                 - LN_NELDERMEAD: The Nelder Mead Simplex algorithm, which seems to perform not so well, but can be used
                   to narrow down a minimum.
                 - LN_SBPLX: SubPlex is an improved version of the Simplex algorithms and usually performs better.

                 With derivative:
                 - LD_MMA: Method of Moving Asymptotes, an improved CCSA
                   ("conservative convex separable approximation") variant of the original MMA algorithm
                 - LD_SLSQP: this is a sequential quadratic programming (SQP) algorithm for (nonlinearly constrained)
                   gradient-based optimization
                 - LD_LBFGS: version of the famous low-storage BFGS algorithm, an approximate Newton method. The same as
                   the Minuit algorithm is built on.
                 - LD_TNEWTON_PRECOND_RESTART, LD_TNEWTON_PRECOND, LD_TNEWTON_RESTART, LD_TNEWTON: a preconditioned
                   inexact truncated Newton algorithm. Multiple variations, with and without preconditioning and/or
                   restart are provided.
                 - LD_VAR1, LD_VAR2: a shifted limited-memory variable-metric algorithm, either using a rank 1 or rank 2
                   method.

             tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
             gradient: Gradient that will be given to the minimizer if supported.
             hessian: Hessian that will be given to the minimizer if supported.
             internal_tols: Tolerances for the minimizer. Has to contain possible tolerance criteria.
             verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
             criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
             strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
             maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
             minimizer_options: Additional options that will be set in the minimizer.
             name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        self._algorithm = algorithm
        if minimizer_options is None:
            minimizer_options = {}
        minimizer_options = copy.copy(minimizer_options)

        if gradient is not NOT_SUPPORTED:
            if gradient is False:
                raise ValueError("grad cannot be False for NLopt minimizer.")
            minimizer_options['gradient'] = gradient
        if hessian is not NOT_SUPPORTED:
            minimizer_options['hessian'] = hessian

        if internal_tols is None:
            internal_tols = {}
        else:
            internal_tols = copy.copy(internal_tols)
        for nlopt_tol in self._ALL_NLOPT_TOL:
            if nlopt_tol not in internal_tols:
                internal_tols[nlopt_tol] = None
        self._internal_tols = internal_tols

        # private kept variables
        self._internal_maxiter = 20
        self._nrandom_max = 5

        super().__init__(name=name,
                         tol=tol,
                         verbosity=verbosity,
                         minimizer_options=minimizer_options,
                         strategy=strategy,
                         criterion=criterion,
                         maxiter=maxiter)

    @minimize_supports(init=True)
    def _minimize(self, loss, params, init):
        previous_result = init
        if init:
            assign_values(params=params, values=init)
        evaluator = self.create_evaluator(loss, params)

        # create minimizer instance
        minimizer = nlopt.opt(nlopt.LD_LBFGS, len(params))

        # initial values as array
        xvalues = np.array(run(params))

        # get and set the limits
        lower = np.array([p.lower for p in params])
        upper = np.array([p.upper for p in params])
        minimizer.set_lower_bounds(lower)
        minimizer.set_upper_bounds(upper)

        # create and set objective function. Either returns only the value or the value
        # and sets the gradient in-place
        def obj_func(x, grad):
            if grad.size > 0:
                value, gradients = evaluator.value_gradient(x)
                grad[:] = np.array(run(gradients))
            else:
                value = evaluator.value(x)

            return value

        minimizer.set_min_objective(obj_func)

        # set maximum number of iterations, also set in evaluator
        minimizer.set_maxeval(self.get_maxiter(len(params)))

        minimizer_options = self.minimizer_options.copy()
        local_minimizer = None
        local_minimizer_options = minimizer_options.pop("local_minimizer_options", None)
        if local_minimizer_options is not None:
            local_minimizer = nlopt.opt(nlopt.LD_LBFGS, len(params))
            local_minimizer.set_lower_bounds(lower)
            local_minimizer.set_upper_bounds(upper)

        maxcor = minimizer_options.pop('maxcor', None)
        if maxcor is not None:
            minimizer.set_vector_storage(maxcor)

        population = minimizer_options.pop('population', None)
        if population is not None:
            minimizer.set_population(population)

        for name, value in minimizer_options.items():
            minimizer.set_param(name, value)

        minimizer.set_param('verbosity', max(0, self.verbosity - 6))

        criterion = self.criterion(tol=self.tol, loss=loss, params=params)
        init_tol = min([math.sqrt(loss.errordef * self.tol), loss.errordef * self.tol * 1e3])
        # init_tol *= 10
        internal_tol = self._internal_tols
        internal_tol = {tol: init_tol if init is None else init for tol, init in internal_tol.items()}
        if 'xtol' in internal_tol:
            internal_tol['xtol'] **= 0.5
        if 'ftol' in internal_tol:
            internal_tol['ftol'] **= 0.5

        valid = None
        edm = None
        criterion_value = None
        maxiter_reached = False
        valid_message = ""
        result_prelim = previous_result
        nrandom = 0
        for i in range(self._internal_maxiter):
            init_scale = []
            approx_step_sizes = {}
            if result_prelim:
                approx_step_sizes = result_prelim.hesse(params=params, method='approx')
            empty_dict = {}
            for param in params:
                step_size = approx_step_sizes.get(param, empty_dict).get('error')
                if step_size is None and param.has_step_size:
                    step_size = param.step_size
                init_scale.append(step_size)

            minimizer.set_initial_step(init_scale)

            self._set_tols_inplace(minimizer=minimizer,
                                   internal_tol=internal_tol,
                                   criterion_value=criterion_value)

            # some (global) optimizers use a local minimizer, set that here
            if local_minimizer is not None:
                self._set_tols_inplace(minimizer=local_minimizer,
                                       internal_tol=internal_tol,
                                       criterion_value=criterion_value)

                minimizer.set_local_optimizer(local_minimizer)

            # run the minimization
            try:
                xvalues = minimizer.optimize(xvalues)
            except MaximumIterationReached:
                maxiter_reached = True
                valid = False
                valid_message = "Maxiter reached, terminated without convergence"
            except RuntimeError:
                if self.verbosity > 3:
                    print("Minimization in NLopt failed, restarting with slightly varied parameters.")
                if nrandom < self._nrandom_max:  # in order not to start too close
                    init_scale = np.where(init_scale != None, init_scale, np.ones_like(init_scale))
                    init_scale_no_nan = np.nan_to_num(init_scale, nan=1.)

                    xvalues += np.random.uniform(low=-init_scale_no_nan, high=init_scale_no_nan) / 2
                    nrandom += 1
            else:
                maxiter_reached = evaluator.niter > evaluator.maxiter

            assign_values(params, xvalues)
            fmin = minimizer.last_optimum_value()  # TODO: what happens if minimization terminated?
            with evaluator.ignore_maxiter():
                result_prelim = FitResult.from_nlopt(loss,
                                                     minimizer=self,
                                                     opt=minimizer,
                                                     edm=CRITERION_NOT_AVAILABLE, niter=evaluator.nfunc_eval,
                                                     params=params,
                                                     evaluator=evaluator,
                                                     criterion=None,
                                                     values=xvalues,
                                                     valid=valid, message=valid_message)
                converged = criterion.converged(result_prelim)
                valid = converged
            criterion_value = criterion.last_value
            if isinstance(criterion, EDM):
                edm = criterion.last_value
            else:
                edm = CRITERION_NOT_AVAILABLE

            if self.verbosity > 5:
                print_minimization_status(converged=converged, criterion=criterion, evaluator=evaluator, i=i, fmin=fmin,
                                          internal_tol=internal_tol)

            if converged or maxiter_reached:
                break

            # update the tols
            self._update_tol_inplace(criterion_value=criterion_value, internal_tol=internal_tol)

        else:
            valid = False
            valid_message = f"Invalid, criterion {criterion.name} is {criterion_value}, target {self.tol} not reached."

        return FitResult.from_nlopt(loss, minimizer=self.copy(), opt=minimizer, edm=edm,
                                    niter=evaluator.niter, params=params, evaluator=evaluator,
                                    values=xvalues, valid=valid, criterion=criterion,
                                    message=valid_message)

    def _set_tols_inplace(self, minimizer, internal_tol, criterion_value):
        # set all the tolerances
        fatol = internal_tol.get('fatol')
        if fatol is not None:
            minimizer.set_ftol_abs(fatol ** 0.5)
        xatol = internal_tol.get('xatol')
        if xatol is not None:
            # minimizer.set_xtol_abs([xatol] * len(params))
            minimizer.set_xtol_abs(xatol)
        # set relative tolerances later as it can be unstable. Just use them when approaching
        if criterion_value is not None:
            tol_factor_full = self.tol / criterion_value
            if tol_factor_full < 1e-8:
                ftol = internal_tol.get('ftol')
                if ftol is not None:
                    minimizer.set_ftol_rel(ftol)

                xtol = internal_tol.get('xtol')
                if xtol is not None:
                    # minimizer.set_xtol_rel([xtol] * len(params))  # TODO: one value or vector?
                    minimizer.set_xtol_rel(xtol)  # TODO: one value or vector?


class NLoptLBFGSV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 maxcor: Optional[int] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt L-BFGS V1"
                 ) -> None:
        """Local, gradient-based quasi-Newton minimizer using the low storage BFGS Hessian approximation.

        This is most probably the most popular algorithm for gradient based local minimum searches and also
        the underlying algorithm in the
        `Minuit <https://www.sciencedirect.com/science/article/abs/pii/0010465575900399>`_ minimizer that is
        also available as :class:`~zfit.minimize.Minuit`.


        This algorithm is based on a Fortran implementation of the low-storage BFGS algorithm
        written by Prof. Ladislav Luksan, and graciously posted online under the GNU LGPL at:

        -   <http://www.uivt.cas.cz/~luksan/subroutines.html>

        The original L-BFGS algorithm, based on variable-metric updates via Strang recurrences,
         was described by the papers:

        -   J. Nocedal, "Updating quasi-Newton matrices with limited storage," *Math. Comput.* **35**, 773-782 (1980).
        -   D. C. Liu and J. Nocedal, "On the limited memory BFGS method for large scale optimization,"
            ''Math. Programming' **45**, p. 503-528 (1989).

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            maxcor: |@doc:minimizer.maxcor| Maximum number of memory history to keep
                   when using a quasi-Newton update formula such as BFGS.
                   It is the number of gradients
                   to “remember” from previous optimization
                   steps: increasing it increases
                   the memory requirements but may speed up the convergence. |@docend:minimizer.maxcor|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LD_LBFGS,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={'maxcor': maxcor},
                         strategy=strategy,
                         maxiter=maxiter)

        @property
        def maxcor(self):
            return self.minimizer_options.get('maxcor')


class NLoptShiftVarV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 maxcor: Optional[int] = None,
                 rank: Optional[int] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name="NLopt Shifted Variable Memory") -> None:
        """Local, gradient-based minimizer using a shifted limited-memory variable-metric.

        This algorithm is based on a Fortran implementation of a shifted limited-memory variable-metric
        algorithm by Prof. Ladislav Luksan, and graciously posted online under the GNU LGPL at:

        -   <http://www.uivt.cas.cz/~luksan/subroutines.html>

        There are two variations of this algorithm: either using a rank-2 method or a rank-1 method.

        The algorithms are based on the ones described by:

        -   J. Vlcek and L. Luksan, "Shifted limited-memory variable metric methods for large-scale unconstrained
            minimization," *J. Computational Appl. Math.* **186**, p. 365-390 (2006).

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:

            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            maxcor: |@doc:minimizer.maxcor| Maximum number of memory history to keep
                   when using a quasi-Newton update formula such as BFGS.
                   It is the number of gradients
                   to “remember” from previous optimization
                   steps: increasing it increases
                   the memory requirements but may speed up the convergence. |@docend:minimizer.maxcor|
            rank: Rank of the algorithm used, either 1 or 2. Defaults to 2.
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """

        if rank is None:
            rank = 2

        if rank == 1:
            algorithm = nlopt.LD_VAR1
        elif rank == 2:
            algorithm = nlopt.LD_VAR2
        else:
            raise ValueError(f"rank has to be either 1 or 2, not {rank}")
        self._rank = rank
        super().__init__(name=name,
                         algorithm=algorithm,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={'maxcor': maxcor},
                         strategy=strategy,
                         maxiter=maxiter)

        @property
        def rank(self):
            return self._rank

        @property
        def maxcor(self):
            return self.minimizer_options.get('maxcor')


class NLoptTruncNewtonV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 maxcor: Optional[int] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name="NLopt Truncated Newton") -> None:
        """Local, gradient-based truncated Newton minimizer using an inexact algorithm.

        This algorithm is based on a Fortran implementation of a
        preconditioned inexact truncated Newton algorithm written by
        Prof. Ladislav Luksan, and graciously posted online under the GNU LGPL
        at:

        -  http://www.uivt.cas.cz/~luksan/subroutines.html

        The algorithms is based on the ones described by:

        -  R. S. Dembo and T. Steihaug, “Truncated Newton algorithms for
           large-scale optimization,” *Math. Programming* **26**, p. 190-212
           (1983) http://doi.org/10.1007/BF02592055.

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            maxcor: |@doc:minimizer.maxcor| Maximum number of memory history to keep
                   when using a quasi-Newton update formula such as BFGS.
                   It is the number of gradients
                   to “remember” from previous optimization
                   steps: increasing it increases
                   the memory requirements but may speed up the convergence. |@docend:minimizer.maxcor|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LD_TNEWTON_PRECOND_RESTART,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={'maxcor': maxcor},
                         strategy=strategy,
                         maxiter=maxiter)

        @property
        def maxcor(self):
            return self.minimizer_options.get('maxcor')


class NLoptSLSQPV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt SLSQP"
                 ) -> None:
        r"""Local gradient based minimizer using a sequential quadratic programming.

        This is a sequential quadratic programming (SQP) algorithm for
        non-linearly gradient-based optimization based on the implementation by
        Dieter Kraft and described in:

        -  Dieter Kraft, “A software package for sequential quadratic
           programming”, Technical Report DFVLR-FB 88-28, Institut für Dynamik
           der Flugsysteme, Oberpfaffenhofen, July 1988.
        -  Dieter Kraft, “Algorithm 733: TOMP–Fortran modules for optimal
           control calculations,” *ACM Transactions on Mathematical Software*,
           vol. 20, no. 3, pp. 262-281 (1994).

        The algorithm optimizes
        successive second-order (quadratic/least-squares) approximations of the
        objective function (via BFGS updates), with first-order (affine)
        approximations of the constraints.

        The Fortran code was obtained from the SciPy project, who are
        responsible for `obtaining permission`_ to distribute it under a
        free-software (3-clause BSD) license.

        The code was modified for inclusion in NLopt by S. G. Johnson in 2010,
        with the following changes. The code was converted to C and manually
        cleaned up. It was modified to be re-entrant (preserving the
        reverse-communication interface but explicitly saving the state in a
        data structure). The inexact line
        search was modified to evaluate the functions including gradients for
        the first step, since this removes the need to evaluate the
        function+gradient a second time for the same point in the common case
        when the inexact line search concludes after a single step.

        **Note:** Because the SLSQP code uses dense-matrix methods (ordinary
        BFGS, not low-storage BFGS), it requires *O*\ (*n*\ 2) storage and
        *O*\ (*n*\ 3) time in *n* dimensions, which makes it less practical for
        optimizing more than a few thousand parameters.

        .. _obtaining permission: http://permalink.gmane.org/gmane.comp.python.scientific.devel/6725

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LD_SLSQP,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={},
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptBOBYQAV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt BOBYQA"
                 ) -> None:
        """Derivative-free local minimizer that iteratively constructed quadratic approximation for the loss.

        This is an algorithm derived from the `BOBYQA subroutine`_ of M. J. D.
        Powell, converted to C and modified for the NLopt stopping criteria.
        BOBYQA performs derivative-free bound-constrained optimization using an
        iteratively constructed quadratic approximation for the objective
        function. See:

        -  M. J. D. Powell, “`The BOBYQA algorithm for bound constrained
           optimization without derivatives`_,” Department of Applied
           Mathematics and Theoretical Physics, Cambridge England, technical
           report NA2009/06 (2009).

        (Because BOBYQA constructs a quadratic approximation of the objective,
        it may perform poorly for objective functions that are not
        twice-differentiable.)

        This algorithm largely
        supersedes the NEWUOA algorithm, which is an earlier version of
        the same idea by Powell.

        .. _BOBYQA subroutine: http://plato.asu.edu/ftp/other_software/bobyqa.zip
        .. _The BOBYQA algorithm for bound constrained optimization without derivatives: http://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LN_BOBYQA,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={},
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptMMAV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt MMA"):
        """Method-of-moving-asymptotes for gradient-based local minimization.

        Globally-convergent method-of-moving-asymptotes (MMA) for gradient-based local minimization.
        The algorithm is described in:

        -  Krister Svanberg, “`A class of globally convergent optimization
           methods based on conservative convex separable approximations`_,”
           *SIAM J. Optim.* **12** (2), p. 555-573 (2002).

        This is an improved CCSA (“conservative convex separable approximation”)
        variant of the original MMA algorithm published by Svanberg in 1987,
        which has become popular for topology optimization. (*Note:* “globally
        convergent” does *not* mean that this algorithm converges to the global
        optimum; it means that it is guaranteed to converge to *some* local
        minimum from any feasible starting point.)

        At each point **x**, MMA forms a local approximation using the gradient
        of *f* and the constraint functions, plus a quadratic “penalty” term to
        make the approximations “conservative” (upper bounds for the exact
        functions). The precise approximation MMA forms is difficult to describe
        in a few words, because it includes nonlinear terms consisting of a
        poles at some distance from *x* (outside of the current trust region),
        almost a kind of Padé approximant. The main point is that the
        approximation is both convex and separable, making it trivial to solve
        the approximate optimization by a dual method. Optimizing the
        approximation leads to a new candidate point **x**. The objective and
        constraints are evaluated at the candidate point. If the approximations
        were indeed conservative (upper bounds for the actual functions at the
        candidate point), then the process is restarted at the new **x**.
        Otherwise, the approximations are made more conservative (by increasing
        the penalty term) and re-optimized.


        .. _A class of globally convergent optimization methods based on conservative convex separable approximations: http://citeseer.ist.psu.edu/viewdoc/summary?doi=10.1.1.146.5196
        .. _Professor Svanberg: http://researchprojects.kth.se/index.php/kb_7902/pb_2085/pb.html

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LD_MMA,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={},
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptCCSAQV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt CCSAQ"):
        """MMA-like minimizer with simpler, quadratic local approximations.

        CCSA is an algorithm from the same paper as MMA as described in:

        -  Krister Svanberg, “`A class of globally convergent optimization
           methods based on conservative convex separable approximations`_,”
           *SIAM J. Optim.* **12** (2), p. 555-573 (2002).

        CCSA has the following differences:
        instead of constructing local MMA approximations, it
        constructs simple quadratic approximations (or rather, affine
        approximations plus a quadratic penalty term to stay conservative). This
        is the ccsa_quadratic code. It seems to have similar convergence rates
        to MMA for most problems, which is not surprising as they are both
        essentially similar.

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LD_CCSAQ,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={},
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptCOBYLAV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt COBYLA"):
        r"""Derivative free simplex minimizer using a linear approximation with trust region steps.

        COBYLA (Constrained Optimization BY Linear Approximations) constructs successive linear approximations of the
        objective function and constraints via a simplex of n+1 points (in n dimensions), and optimizes these
        approximations in a trust region at each step.

        This is a derivative of Powell’s implementation of the COBYLA  algorithm for derivative-free optimization
        by M. J. D. Powell described in:

        -  M. J. D. Powell, “A direct search optimization method that models the
           objective and constraint functions by linear interpolation,” in
           *Advances in Optimization and Numerical Analysis*, eds. S. Gomez and
           J.-P. Hennart (Kluwer Academic: Dordrecht, 1994), p. 51-67.

        and reviewed in:

        -  M. J. D. Powell, “Direct search algorithms for optimization
           calculations,” *Acta Numerica* **7**, 287-336 (1998).

        It constructs successive linear approximations of the objective function
        and constraints via a simplex of *n*\ +1 points (in *n* dimensions), and
        optimizes these approximations in a trust region at each step.

        The original code itself was written in Fortran by Powell and was
        converted to C in 2004 by Jean-Sebastien Roy (js@jeannot.org) for the
        SciPy project. The version in NLopt was based on Roy’s C version and offers a few improvements
        over the original code:

        - `COBYLA` can increase the trust-region radius if the predicted improvement
          was approximately right and the simplex is OK
        - pseudo-randomization of the simplex steps in the COBYLA algorithm improve the robustness by avoiding
           accidentally taking steps that don't improve conditioning.

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|


        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LN_COBYLA,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={},
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptSubplexV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt Subplex"):
        """Local derivative free minimizer which improves on the Nealder-Mead algorithm.

        This is a re-implementation of Tom Rowan’s “Subplex” algorithm.

        Subplex (a variant of Nelder-Mead that uses Nelder-Mead on a sequence of
        subspaces) is claimed to be much more efficient and robust than the
        original Nelder-Mead, while retaining the latter’s facility with
        discontinuous objectives.

        The description of Rowan’s algorithm in his PhD thesis is used:

        -  T. Rowan, “Functional Stability Analysis of Numerical Algorithms”,
           Ph.D. thesis, Department of Computer Sciences, University of Texas at
           Austin, 1990.

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|


        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        super().__init__(name=name,
                         algorithm=nlopt.LN_SBPLX,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options={},
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptMLSLV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 population: Optional[int] = None,
                 randomized: Optional[bool] = None,
                 local_minimizer: Optional[Union[int, Mapping[str, object]]] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt MLSL",
                 ) -> None:
        """Global minimizer using local optimization by randomly selecting points.

        “Multi-Level Single-Linkage” (MLSL) is an algorithm for global optimization by
        a sequence of local optimizations from random starting points, proposed
        by:

        -  A. H. G. Rinnooy Kan and G. T. Timmer, “Stochastic global
           optimization methods,” *Mathematical Programming*, vol. 39, p. 27-78
           (1987). (Actually 2 papers — part I: clustering methods, p. 27, then
           part II: multilevel methods, p. 57.)

        We also include a modification of MLSL use a Sobol’ `low-discrepancy
        sequence`_ (LDS), also used in so-called
        `quasi Monte Carlo methods <https://en.wikipedia.org/wiki/Quasi-Monte_Carlo_method>`_
        that can be invoked by setting *randomized* to False
        (as it is now *quasi*randomized) instead of pseudorandom numbers, which was argued to
        improve the convergence rate by:

        -  Sergei Kucherenko and Yury Sytsko, “Application of deterministic
           low-discrepancy sequences in global optimization,” *Computational
           Optimization and Applications*, vol. 30, p. 297-318 (2005).

        In either case, MLSL is a “multistart” algorithm: it works by doing a
        sequence of local optimizations (using some other local optimization
        algorithm) from random or low-discrepancy starting points. MLSL is
        distinguished, however by a “clustering” heuristic that helps it to
        avoid repeated searches of the same local optima, and has some
        theoretical guarantees of finding all local optima in a finite number of
        local minimizations.

        The local-search portion of MLSL can use any of the other algorithms in
        NLopt. The local search uses the
        derivative/nonderivative algorithm set by the *local_minimizer* argument.


        .. _low-discrepancy sequence: https://en.wikipedia.org/wiki/Low-discrepancy_sequence
        .. _local optimization: NLopt_Reference#localsubsidiary-optimization-algorithm

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            population: |@doc:minimizer.nlopt.population| The population size for the evolutionary algorithm. |@docend:minimizer.nlopt.population|
                By default, each iteration of MLSL samples 4 random new trial points.
            randomized: If True, uses the randomized version 'GD_MLSL_LDS' instead of 'GD_MLSL'
            local_minimizer: Configuration for the local minimizer. Defaults to L-BFGS.
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """
        if randomized is None:
            randomized = False
        if randomized:
            algorithm = nlopt.GD_MLSL
        else:
            algorithm = nlopt.GD_MLSL_LDS

        local_minimizer = nlopt.LD_LBFGS if local_minimizer is None else local_minimizer
        if not isinstance(local_minimizer, collections.Mapping):
            local_minimizer = {'algorithm': local_minimizer}
        if 'algorithm' not in local_minimizer:
            raise ValueError("algorithm needs to be specified in local_minimizer")

        minimizer_options = {'local_minimizer_options': local_minimizer}
        if population is not None:
            minimizer_options['population'] = population
        super().__init__(name=name,
                         algorithm=algorithm,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options=minimizer_options,
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptStoGOV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 randomized: Optional[bool] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt MLSL"):
        """Global minimizer which divides the space into smaller rectangles and uses a local BFGS variant inside.

        StoGO is a global optimization algorithm that works by systematically dividing
        the search space (which must be bound-constrained) into smaller
        hyper-rectangles via a branch-and-bound technique, and searching them by
        a gradient-based local-search algorithm (a BFGS variant), optionally
        including some randomness (hence the “Sto”, which stands for
        “stochastic” I believe).

        Some references on StoGO are:

        -  S. Gudmundsson, “Parallel Global Optimization,” M.Sc. Thesis, IMM,
           Technical University of Denmark, 1998.
        -  K. Madsen, S. Zertchaninov, and A. Zilinskas, “Global Optimization
           using Branch-and-Bound,” unpublished (1998). A preprint of this paper
           is included in the ``stogo`` subdirectory of NLopt as ``paper.pdf``.
        -  S. Zertchaninov and K. Madsen, “A C++ Programme for Global
           Optimization,” IMM-REP-1998-04, Department of Mathematical Modelling,
           Technical University of Denmark, DK-2800 Lyngby, Denmark, 1998. A
           copy of this report is included in the ``stogo`` subdirectory of
           NLopt as ``techreport.pdf``.


        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            randomized: If True, uses the randomized version 'GD_STOGO_RAND' instead of 'GD_STOGO'
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """

        if randomized is None:
            randomized = False

        if randomized:
            algorithm = nlopt.GD_STOGO_RAND
        else:
            algorithm = nlopt.GD_STOGO
        super().__init__(name=name,
                         algorithm=algorithm,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options=None,
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptESCHV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt ESCH"):
        """Global minimizer using an evolutionary algorithm.

        This is a modified Evolutionary Algorithm for global optimization,
        developed by Carlos Henrique da Silva Santos’s and described in the
        following paper and Ph.D thesis:

        -  C. H. da Silva Santos, M. S. Gonçalves, and H. E. Hernandez-Figueroa,
           “Designing Novel Photonic Devices by Bio-Inspired Computing,” *IEEE
           Photonics Technology Letters* **22** (15), pp. 1177–1179 (2010).

        .. raw:: html

           <!-- -->

        -  C. H. da Silva Santos, “`Parallel and Bio-Inspired Computing Applied
           to Analyze Microwave and Photonic Metamaterial Strucutures`_,”
           Ph.D. thesis, University of Campinas, (2010).

        The algorithm is adapted from ideas described in:

        -  H.-G. Beyer and H.-P. Schwefel, “Evolution Strategies: A
           Comprehensive Introduction,” *Journal Natural Computing*, **1** (1),
           pp. 3–52 (2002_.

        .. raw:: html

           <!-- -->

        -  Ingo Rechenberg, “Evolutionsstrategie – Optimierung technischer
           Systeme nach Prinzipien der biologischen Evolution,” Ph.D. thesis
           (1971), Reprinted by Fromman-Holzboog (1973).

        .. _Parallel and Bio-Inspired Computing Applied to Analyze Microwave and Photonic Metamaterial Strucutures: http://www.bibliotecadigital.unicamp.br/document/?code=000767537&opt=4&lg=en_US

        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

        Args:
            tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
            verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
            maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
            strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
            criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
            name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """

        algorithm = nlopt.GN_ESCH

        super().__init__(name=name,
                         algorithm=algorithm,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options=None,
                         strategy=strategy,
                         maxiter=maxiter)


class NLoptISRESV1(NLoptBaseMinimizerV1):
    def __init__(self,
                 tol: Optional[float] = None,
                 population: Optional[int] = None,
                 verbosity: Optional[int] = None,
                 maxiter: Optional[Union[int, str]] = None,
                 strategy: Optional[ZfitStrategy] = None,
                 criterion: Optional[ConvergenceCriterion] = None,
                 name: str = "NLopt ISRES"):
        """Improved Stochastic Ranking Evolution Strategy using a mutation rule and differential variation.

         The evolution strategy is based on a combination of a mutation rule (with a log-normal step-size update and
         exponential smoothing) and differential variation (a Nelder–Mead-like update rule).
         The fitness ranking is simply via the objective function for problems without nonlinear constraints,
         but when nonlinear constraints are included the stochastic ranking proposed by Runarsson and Yao is employed.

         The NLopt implementation is based on the method described in:

         - Thomas Philip Runarsson and Xin Yao,
          `"Search biases in constrained evolutionary optimization <http://www3.hi.is/~tpr/papers/RuYa05.pdf>"`_,
          *IEEE Trans. on Systems, Man, and Cybernetics Part C: Applications and Reviews*,
           vol. 35 (no. 2), pp. 233-243 (2005).

         It is a refinement of an earlier method described in:

         - Thomas P. Runarsson and Xin Yao,
           `"Stochastic ranking for constrained evolutionary optimization
           <http://www3.hi.is/~tpr/software/sres/Tec311r.pdf>"`_, *IEEE Trans. Evolutionary Computation*,
           vol. 4 (no. 3), pp. 284-294 (2000).

         The actual implementation is independent provided by S. G. Johnson (2009) based on the papers above.
         Runarsson also has his own Matlab implemention available from `his web page <http://www3.hi.is/~tpr>`_.


        |@doc:minimizer.nlopt.info| More information on the algorithm can be found
        `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_.

        This implenemtation uses internally the
        `NLopt library <https://nlopt.readthedocs.io/en/latest/>`_.
        It is a
        free/open-source library for nonlinear optimization,
        providing a common interface for a number of
        different free optimization routines available online as well as
        original implementations of various other algorithms. |@docend:minimizer.nlopt.info|

         Args:
             tol: |@doc:minimizer.tol| Termination value for the
                   convergence/stopping criterion of the algorithm
                   in order to determine if the minimum has
                   been found. Defaults to 1e-3. |@docend:minimizer.tol|
             population: |@doc:minimizer.nlopt.population| The population size for the evolutionary algorithm. |@docend:minimizer.nlopt.population| Defaults to 20×(n+1) in n dimensions.
             verbosity: |@doc:minimizer.verbosity| Verbosity of the minimizer. Has to be between 0 and 10.
              The verbosity has the meaning:

               - a value of 0 means quiet and no output
               - above 0 up to 5, information that is good to know but without
                 flooding the user, corresponding to a "INFO" level.
               - A value above 5 starts printing out considerably more and
                 is used more for debugging purposes.
               - Setting the verbosity to 10 will print out every
                 evaluation of the loss function and gradient.

               Some minimizer offer additional output which is also
               distributed as above but may duplicate certain printed values. |@docend:minimizer.verbosity|
             maxiter: |@doc:minimizer.maxiter| Approximate number of iterations.
                   This corresponds to roughly the maximum number of
                   evaluations of the `value`, 'gradient` or `hessian`. |@docend:minimizer.maxiter|
             strategy: |@doc:minimizer.strategy| A class of type `ZfitStrategy` that takes no
                   input arguments in the init. Determines the behavior of the minimizer in
                   certain situations, most notably when encountering
                   NaNs. It can also implement a callback function. |@docend:minimizer.strategy|
             criterion: |@doc:minimizer.criterion| Criterion of the minimum. This is an
                   estimated measure for the distance to the
                   minimum and can include the relative
                   or absolute changes of the parameters,
                   function value, gradients and more.
                   If the value of the criterion is smaller
                   than ``loss.errordef * tol``, the algorithm
                   stopps and it is assumed that the minimum
                   has been found. |@docend:minimizer.criterion|
             name: |@doc:minimizer.name| Human readable name of the minimizer. |@docend:minimizer.name|
        """

        algorithm = nlopt.GN_ISRES
        if population is not None:
            minimizer_options = {'population': population}
        else:
            minimizer_options = None

        super().__init__(name=name,
                         algorithm=algorithm,
                         tol=tol,
                         gradient=NOT_SUPPORTED,
                         hessian=NOT_SUPPORTED,
                         criterion=criterion,
                         verbosity=verbosity,
                         minimizer_options=minimizer_options,
                         strategy=strategy,
                         maxiter=maxiter)



#  Copyright (c) 2021 zfit

import numpy as np
import tensorflow as tf

from zfit import z


def numpy_func(x, a):
    return np.square(x) * a


@z.function
def wrapped_numpy_func(x_tensor, a_tensor):
    result = tf.numpy_function(func=numpy_func, inp=[x_tensor, a_tensor], Tout=tf.float64)
    result = tf.sqrt(result)
    return result


def test_wrapped_func():
    rnd = z.random.uniform(shape=(10,))
    result = wrapped_numpy_func(rnd, z.constant(3.))
    np.testing.assert_allclose(rnd * np.sqrt(3), result)



#  Copyright (c) 2021 zfit
import functools
from typing import Mapping, Optional, Set

import tensorflow as tf

import zfit.z.numpy as znp

from .. import z
from ..core.interfaces import ZfitIndependentParameter, ZfitPDF, ZfitSpace
from ..core.parameter import set_values
from ..core.space import combine_spaces, convert_to_space
from ..util.exception import WorkInProgressError
from ..util.warnings import warn_experimental_feature
from .functor import BaseFunctor


class ConditionalPDFV1(BaseFunctor):

    @warn_experimental_feature
    def __init__(self,
                 pdf: ZfitPDF,
                 cond: Mapping[ZfitIndependentParameter, ZfitSpace],
                 name: str = "ConditionalPDF",
                 *,
                 use_vectorized_map: bool = False,
                 sample_with_replacement: bool = True
                 ) -> None:
        """EXPERIMENTAL! Implementation of a Conditional PDF, rather slow and for research purpose.

        As an example, a Gaussian is wrapped in order to make 'sigma' conditional.

        .. jupyter-execute::



        Args:
            pdf: PDF that will be wrapped. Convert one or several parameters of *pdf* to a conditional
                parameter, meaning that the parameter *param* in the ``cond`` mapping will now be
                determined by the data in the ``Space``, the value of the ``cond``.
            cond: Mapping of parameter to input data.
            name: |@doc:model.init.name| Name of the PDF for a better visibility.
               Has no programmatical functional purpose. |@docend:model.init.name|
            use_vectorized_map ():
            sample_with_replacement ():
        """
        self._sample_with_replacement = sample_with_replacement
        self._use_vectorized_map = use_vectorized_map
        self._cond, cond_obs = self._check_input_cond(cond)
        obs = pdf.space * cond_obs
        super().__init__(pdfs=pdf, obs=obs, name=name)
        self.set_norm_range(pdf.norm_range)

    def _check_input_cond(self, cond):
        spaces = []
        for param, obs in cond.items():
            if not isinstance(param, ZfitIndependentParameter):
                raise TypeError(f"parameter {param} not a ZfitIndependentParameter")
            spaces.append(convert_to_space(obs))
        return cond, combine_spaces(*spaces)

    @z.function(wraps='conditional_pdf')
    def _pdf(self, x, norm_range):
        pdf = self.pdfs[0]
        param_x_indices = {p: x.obs.index(p_space.obs[0]) for p, p_space in self._cond.items()}
        x_values = x.value()

        if self._use_vectorized_map:
            tf_map = tf.vectorized_map
        else:
            output_signature = tf.TensorSpec(shape=(1, *x_values.shape[1:-1]), dtype=self.dtype)
            tf_map = functools.partial(tf.map_fn, fn_output_signature=output_signature)

        # TODO: reset parameters?

        def eval_pdf(cond_and_data):
            x_pdf = cond_and_data[None, ..., :pdf.n_obs]
            for param, index in param_x_indices.items():
                param.assign(cond_and_data[..., index])
            return pdf.pdf(x_pdf, norm_range=norm_range)

        params = tuple(param_x_indices.keys())
        with set_values(params, params):
            probs = tf_map(eval_pdf, x_values)
        probs = probs[:, 0]  # removing stack dimension, implicitly in map_fn
        return probs

    def _get_params(self, floating: Optional[bool] = True, is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set["ZfitParameter"]:
        params = super()._get_params(floating, is_yield, extract_independent)
        params -= set(self._cond)
        return params

    @z.function(wraps='conditional_pdf')
    def _single_hook_integrate(self, limits, norm_range, x):

        param_x_indices = {p: x.obs.index(p_space.obs[0]) for p, p_space in self._cond.items()}
        x_values = x.value()
        pdf = self.pdfs[0]

        if self._use_vectorized_map:
            tf_map = tf.vectorized_map
        else:
            output_signature = tf.TensorSpec(shape=(1, *x_values.shape[1:-1]), dtype=self.dtype)
            tf_map = functools.partial(tf.map_fn, fn_output_signature=output_signature)

        def eval_int(values):
            for param, index in param_x_indices.items():
                param.assign(values[..., index])

            return pdf.integrate(limits=limits, norm_range=norm_range, x=x)

        integrals = tf_map(eval_int, x_values)
        integrals = integrals[:, 0]  # removing stack dimension, implicitly in map_fn
        return integrals

    @z.function(wraps='conditional_pdf')
    def _single_hook_sample(self, n, limits, x):
        tf.assert_equal(n, x.nevents, message="Different number of n requested than x given for "
                                              "conditional sampling. Needs to agree")

        param_x_indices = {p: x.obs.index(p_space.obs[0]) for p, p_space in self._cond.items()}
        x_values = x.value()
        # if self._sample_with_replacement:
        #     x_values = z.random.sample_with_replacement(x_values, axis=0, sample_shape=(n,))
        pdf = self.pdfs[0]

        if self._use_vectorized_map:
            tf_map = tf.vectorized_map
        else:
            output_signature = tf.TensorSpec(shape=(1, pdf.n_obs), dtype=self.dtype)
            tf_map = functools.partial(tf.map_fn, fn_output_signature=output_signature)

        def eval_sample(values):
            for param, index in param_x_indices.items():
                param.assign(values[..., index])

            return pdf.sample(n=1, limits=limits).value()

        sample_rnd = tf_map(eval_sample, x_values)[..., 0]
        sample = znp.concatenate([sample_rnd, x_values], axis=-1)
        return sample

    def copy(self, **override_parameters) -> 'BasePDF':
        raise WorkInProgressError("Currently copying not possible. "
                                  "Use `set_yield` to set a yield inplace.")



#  Copyright (c) 2021 zfit

import tensorflow as tf

from .core.constraint import GaussianConstraint, SimpleConstraint
from .util import ztyping

__all__ = ["nll_gaussian", "SimpleConstraint", "GaussianConstraint"]


def nll_gaussian(params: ztyping.ParamTypeInput, observation: ztyping.NumericalScalarType,
                 uncertainty: ztyping.NumericalScalarType) -> tf.Tensor:
    """Return negative log likelihood graph for gaussian constraints on a list of parameters.

    Args:
        params: The parameters to constraint.
        observation: observed values of the parameter.
        uncertainty: Uncertainties or covariance/error.
            matrix of the observed values. Can either be a single value, a list of values, an array or a tensor.
    Returns:
        The constraint object.
    Raises:
        ShapeIncompatibleError: if params, mu and sigma don't have the same size.
    """

    return GaussianConstraint(params=params, observation=observation, uncertainty=uncertainty)
